#include "list.h"

static ls_list *merge(ls_list *, ls_list *, int (*)(const ls_list *, const
ls_list *));

ls_list *
ls_new(size_t sz, void (*f)(ls_list *)) {
	ls_list *new;
	new = malloc(sz);
	if (new) {
		new->next = NULL;
		if (f)
			f(new);
	}
	return new;
}

void
ls_free(ls_list *v, void (*f)(ls_list *)) {
	ls_list *mp;

	while (v) {
		mp = v->next;
		if (f)
			f(v);
		free(v);
		v = mp;
	}
}

ls_list *
ls_append(ls_list *dest, ls_list *src) {
	src->next = dest->next;
	dest->next = src;
	return dest;
}

ls_list *
ls_push(ls_list *dest, ls_list *src) {

	src->next = dest;
	return src;
}

ls_list *
ls_iter(ls_list *v, int (*f)(ls_list *)) {
	while (v) {
		if (f && f(v))
			return v;
		v = v->next;
	}
	return v;
}

ls_list *
ls_sort(ls_list *source, int nitems,
	int (*cmp)(const ls_list *, const ls_list *)) {
	ls_list *lp, *end, *mid;
	int i;

	if (!source || !nitems)
		return 0;

	if (nitems < 0) {
		i = 0;
		for (lp = source; lp; lp = lp->next)
			++i;
		nitems = i;
	}

	lp = source;

	/* special cases we do without merging */
	switch (nitems) {
	case 1:
		return source;
		break;
	case 2:
		/* if list is out of order... */
		if (cmp(lp, lp->next) > 0) {
			end = lp->next;
			end->next = lp;
			lp->next = 0;
			return end;
		} else {
			return lp;
		}
		break;
	case 3:
		end = lp->next->next;
		lp->next->next = 0;

		if (cmp(lp, lp->next) > 0) {
			mid = lp->next;
			mid->next = lp;
			lp->next = 0;
		} else {
			mid = lp;
		}
		return merge(mid, end, cmp);
		break;
	}

	i = nitems / 2;

	while (--i)
		lp = lp->next;

	i = nitems / 2;
	mid = lp->next;
	lp->next = 0;

	return merge(ls_sort(source, i, cmp),
		     ls_sort(mid, nitems - i, cmp), cmp);
}

static ls_list *
merge(ls_list *lp1, ls_list *lp2,
	int (*cmp)(const ls_list *, const ls_list *)) {
	ls_list *out, *walk;

	if (!lp1)
		return lp2;
	if (!lp2)
		return lp1;

	if (cmp(lp1, lp2) > 0) {
		out = lp2;
		lp2 = lp2->next;
	} else {
		out = lp1;
		lp1 = lp1->next;
	}

	walk = out;

	while (lp1 && lp2) {
		while (lp1 && cmp(lp1, lp2) < 1) {
			walk->next = lp1;
			walk = lp1;
			lp1 = lp1->next;
		}
		if (!lp1)
			break;
		while (lp2 && cmp(lp2, lp1) < 0) {
			walk->next = lp2;
			walk = lp2;
			lp2 = lp2->next;
		}
	}

	if (lp1) {
		walk->next = lp1;
	} else {
		walk->next = lp2;
	}
	/* and it's already NULL terminated. */

	return out;
}

ls_list *
ls_rev(ls_list *lp) {
	ls_list *walk, *ret;

	ret = NULL;

	while (lp) {
		walk = lp->next;
		lp->next = ret;
		ret = lp;
		lp = walk;
	}

	return ret;
}
